# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dimod import SampleSet, BinaryQuadraticModel
from typing import Callable

from .termination import Terminator, Termination


class LowestEnergyTerminator(Terminator):
    """
    Requests termination, if a `predicate` is satisfied by the current
    populations's lowest energy.

    """
    def __init__(self, predicate: Callable[[float], bool]):
        self._predicate = predicate

    def should_terminate(self, population: SampleSet) -> bool:
        return self._predicate(population.first.energy)



class LowestEnergyTermination(Termination):
    """
    A factory that produces instances `LowestEnergyTerminator`.
    """
    def __init__(self, predicate: Callable[[float], bool]):
        self._predicate = predicate

    def initialise(self, problem: BinaryQuadraticModel) -> Terminator:
        return LowestEnergyTerminator(self._predicate)



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from dimod import SampleSet, Vartype, BinaryQuadraticModel, as_samples
from numpy.random import Generator

from .startup import Startup

MAX_ITERATIONS = 1_000

def create_random(num_qubits: int, size: int, vartype: Vartype, max_tries: int, rng: Generator) -> np.ndarray:
    assert num_qubits > 0
    assert size > 0

    choosable = np.array(list(vartype.value), dtype='i1')

    sample = np.unique(rng.choice(choosable, size=(size, num_qubits)), axis=0)
    tries = 1

    while len(sample) != size and tries <= max_tries:
        additional = rng.choice(choosable, size=(size-len(sample), num_qubits))
        sample = np.unique(np.concatenate([sample, additional]), axis=0)
        tries += 1

    return sample


class RandomStartup(Startup):
    def __init__(self, population_size: int, rng: np.random.Generator):
        assert population_size > 1
        self._population_size = population_size
        self._rng = rng

    def startup(self, problem: BinaryQuadraticModel) -> SampleSet:
        sample = create_random(len(problem), self._population_size, problem.vartype, MAX_ITERATIONS, self._rng)
        return SampleSet.from_samples_bqm((sample, problem.variables), problem, aggregate_samples=True)



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from numpy.random import Generator
from dimod import SampleSet, BinaryQuadraticModel
from typing import Callable

from .recombination import Recombinator
from .pair_recombination import PairRecombinator, PairRecombination


class RandomBitsRecombinator(PairRecombinator):
    """
    Randomly flips differing bits for randomly created pairs.

    :ivar bqm: The problem to optimise.
    :ivar num_bits: Number of bits to flip.
    :ivars recombination_rate: Amount of mates each individual should be with paired.
    :ivars rng: Random number generator for reproduciblility.
    """
    def __init__(self, bqm: BinaryQuadraticModel, num_bits: int, recombination_rate: int, rng: np.random.Generator):
        self._bqm = bqm
        self._num_bits = num_bits
        super().__init__(recombination_rate, rng)

    def _mate(self, population: SampleSet, pairs: np.ndarray) -> SampleSet:
        strings = population.record.sample
        left = strings[pairs[0]]
        right = strings[pairs[1]]

        result = np.zeros((2 * len(pairs), strings.shape[1]), dtype=strings.dtype)

        for i in range(len(pairs)):
            arg_xor = np.flatnonzero(left[i] != right[i])
            num_bits = min(self._num_bits, len(arg_xor))
            bits = self._rng.choice(arg_xor, size=num_bits)

            result[i] = left[i]
            result[i][bits] = right[i][bits]

            j = 2 * i
            result[j] = right[i]
            result[j][bits] = left[i][bits]

        return SampleSet.from_samples_bqm((result, self._bqm.variables), self._bqm, aggregate_samples=True)


class RandomBitsRecombination(PairRecombination):
    """
    A factory that produces instances `RandomBitsRecombination`.
    """
    def __init__(self, num_bits: int, recombination_rate: int, rng: Callable[[], Generator]):
        self._num_bits = num_bits
        super().__init__(recombination_rate, rng)

    def initialise(self, problem: BinaryQuadraticModel) -> Recombinator:
        return RandomBitsRecombinator(problem, self._num_bits, self._recombination_rate, self._rng())



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from dimod import SampleSet, BinaryQuadraticModel

from .recombination import Recombinator
from .pair_recombination import PairRecombinator, PairRecombination


class RandomClusterRecombinator(PairRecombinator):
    """
    Randomly chooses coherent regions of differing bits and flips them
    for randomly created pairs.

    :ivar bqm: The problem to optimise.
    :ivars recombination_rate: Amount of mates each individual should be with paired.
    :ivars rng: Random number generator for reproduciblility.
    """
    def __init__(self, bqm: BinaryQuadraticModel, recombination_rate: int, rng: np.random.Generator):
        self._bqm = bqm
        super().__init__(recombination_rate, rng)

    def _mate(self, population: SampleSet, pairs: np.ndarray) -> SampleSet:
        strings = population.record.sample
        left = strings[pairs[0]]
        right = strings[pairs[1]]

        result = np.zeros((2 * len(pairs), strings.shape[1]), dtype=strings.dtype)

        for i in range(len(pairs)):
            arg_xor = np.flatnonzero(left[i] != right[i])
            bit = self._rng.choice(arg_xor, size=1)[0]

            j = 2 * i
            result[i] = left[i]
            result[j] = right[i]

            cluster_start = bit
            while cluster_start >= 0 and right[i][cluster_start] != left[i][cluster_start]:
                result[i][cluster_start] = right[i][cluster_start]
                result[j][cluster_start] = left[i][cluster_start]
                cluster_start -= 1

            cluster_end = bit
            while cluster_end < strings.shape[1] and right[i][cluster_end] != left[i][cluster_end]:
                result[i][cluster_end] = right[i][cluster_end]
                result[j][cluster_end] = left[i][cluster_end]
                cluster_end += 1


        return SampleSet.from_samples_bqm((result, self._bqm.variables), self._bqm, aggregate_samples=True)


class RandomClusterRecombination(PairRecombination):
    """
    A factory that produces instances `RandomClusterRecombination`.
    """
    def initialise(self, problem: BinaryQuadraticModel) -> Recombinator:
        return RandomClusterRecombinator(problem, self._recombination_rate, self._rng())



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dimod import SampleSet, BinaryQuadraticModel
from typing import List

from .termination import Terminator, Termination


class OrTerminator(Terminator):
    """
    Combines multiple instances of `Terminator` and requests termination
    if at least one of the children do.

    :ivar stoppers: Children to combine
    """
    def __init__(self, stoppers: List[Terminator]):
        self.stoppers = stoppers

    def should_terminate(self, population: SampleSet) -> bool:
        return any(s.should_terminate(population) for s in self.stoppers)


class OrTermination(Termination):
    """
    A factory that produces instances `OrTerminator`.
    """
    def __init__(self, stops: List[Termination]):
        self.stops = stops

    def initialise(self, problem: BinaryQuadraticModel) -> Terminator:
        return OrTerminator([s.initialise(problem) for s in self.stops])



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod

from dimod import SampleSet, BinaryQuadraticModel


class Recombinator(ABC):
    """
    An interface for improving solution candidates by combining them.

    Instances of subclasses are usually created once by a suitable :class:`Recombinator` factory at the start of an
    optimisation run. In each iteration step :meth:`recombine` is called to assign individuals of the current
    population to mating partners. Then offsprings are procreated by finding bits in which the parents differ and
    swapping them.

    When iteration terminates, the :class:`Recombinator` is destroyed. Thus, each run should have its own
    :class:`Recombinator` that is independent from any other.
    """
    @abstractmethod
    def recombine(self, population: SampleSet) -> SampleSet:
        """
        Creates improved solution candidates by combining individuals of the calling algorithm's current population.

        Implementations of this interface must ensure that in the result population no solution appears twice.
        Moreover, the best individual of the new population must be at least as good as the best one of the old population.

        :param population: The calling algorithm's current population.
        """
        pass


class Recombination(ABC):
    """
    An interface that is used to configure, how to combine individuals to procreate enhanced ones per iteration.

    In fact, it is a factory that produces instances of the interface :class:`Recombinator`.
    """
    @abstractmethod
    def initialise(self, problem: BinaryQuadraticModel) -> Recombinator:
        """
        Instantiates and configures a implementors of the interface `Recombinator`.

        :param problem: The problem that should be optimised in the current run.
        """
        pass



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from abc import abstractmethod
from numpy.random import Generator, default_rng
from typing import Callable

from dimod import SampleSet

from .recombination import Recombinator, Recombination



def create_pairs(population_size: int, recombination_rate: int, rng: Generator = default_rng()):
    """
    Tries to choose randomly `recombination_rate` mates for each element in population_size.
    An element cannot be mated with itself and there are no dublicate pairs.

    :param population_size: Number of individuals that should be mated.
    :param recombination_rate: Amount of mates each individual should be with paired.
    :param rng: Random number generator.
    """
    assert population_size >= 2
    assert recombination_rate >= 1

    # If recombination rate is too large, create all possible pairs
    if population_size <= recombination_rate:
        recombination_rate = population_size - 1


    # Stores pairs
    assignment = np.zeros((2, population_size * recombination_rate), dtype=int)
    # Stores which mates are choosable in the respective iteration
    assignable = np.arange(1, population_size, dtype=int)

    # Sample mates for first individual
    assignment[1, :recombination_rate] = rng.choice(assignable, size=recombination_rate, replace=False)

    chosen = recombination_rate
    for individual in range(1, population_size):
        # Prohibit choosing one self
        assignable[individual-1] -= 1

        # Sample mates for current individual
        new_chosen = chosen + recombination_rate
        assignment[0, chosen:new_chosen] = individual
        assignment[1, chosen:new_chosen] = rng.choice(assignable, size=recombination_rate, replace=False)

        chosen = new_chosen

    # Eliminate duplicates
    assignment = np.unique(np.sort(assignment, axis=0), axis=1)

    return assignment


class PairRecombinator(Recombinator):
    """
    An auxiliary class for instances of `Recombinator` that combine pairs of individuals.
    Their random creation is implemented here.

    :ivars population_size: Number of individuals that should be mated.
    :ivars recombination_rate: Amount of mates each individual should be with paired.
    :ivars rng: Random number generator.
    """
    def __init__(self, recombination_rate: int, rng: np.random.Generator):
        assert recombination_rate > 1
        self._recombination_rate = recombination_rate
        self._rng = rng

    def recombine(self, population: SampleSet) -> SampleSet:
        pairs = create_pairs(len(population), self._recombination_rate)
        return self._mate(population, pairs)

    @abstractmethod
    def _mate(self, population: SampleSet, pairs: np.ndarray) -> SampleSet:
        pass


class PairRecombination(Recombination):
    """
    A factory that produces instances `PairRecombination`.
    """
    def __init__(self, recombination_rate: int, rng: Callable[[], np.random.Generator]):
        assert recombination_rate > 1
        self._recombination_rate = recombination_rate
        self._rng = rng



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from dimod import BinaryQuadraticModel
from dimod.sampleset import SampleSet, concatenate

from .recombination import Recombinator, Recombination

class AdditionRecombinator(Recombinator):
    """
    Executes a wrapped `Recombinator` and merges its result with the old population.
    This is a trivial way of preserving good solution quality.

    :ivars internal: Defines how new individuals are appended to current population.
    """
    def __init__(self, internal: Recombinator):
        self._internal = internal

    def recombine(self, population: SampleSet) -> SampleSet:
        return concatenate([population, self._internal.recombine(population)]).aggregate()


class AdditionRecombination(Recombination):
    """
    A factory that produces instances `AdditionRecombination`.
    """
    def __init__(self, internal: Recombination):
        self._internal = internal

    def initialise(self, problem: BinaryQuadraticModel) -> Recombinator:
        return AdditionRecombinator(self._internal.initialise(problem))



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dimod import SampleSet, BinaryQuadraticModel

from .selection import Selector, Selection


class TruncationSelector(Selector):
    """
    Selects the best `num_preserved` individuals, i.e. those with the lowest
    energies.

    """
    def __init__(self, num_preserved: int):
        assert num_preserved > 1
        self._num_preserved = num_preserved

    def select(self, population: SampleSet) -> SampleSet:
        return population.truncate(self._num_preserved)


class TruncationSelection(Selection):
    """
    A factory that produces instances `TruncationSelection`.
    """
    def __init__(self, num_preserved: int):
        assert num_preserved > 1
        self._num_preserved = num_preserved

    def initialise(self, problem: BinaryQuadraticModel) -> Selector:
        return TruncationSelector(self._num_preserved)



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dimod import Sampler, BinaryQuadraticModel
from typing import Callable, Dict, Any
from numpy.random import Generator

from qaga import mutation as _mutation
from qaga import recombination as _recombination
from qaga import selection as _selection
from qaga import termination as _termination

from .genetic_algorithm import GeneticAlgorithm
from .startup import RandomStartup, WarmStartup


class IdentityMutation(_mutation.IdentityMutation):
    """
    Passes the input population without any changes.
    """
    pass


class IdentityRecombination(_recombination.IdentityRecombination):
    """
    Passes the input population without any changes.
    """
    pass


class IdentitySelection(_selection.IdentitySelection):
    """
    Passes the input population without any changes.
    """
    pass


class ReverseAnnealingMutation(_mutation.Mutation):
    """
    Improves each individual by running a `Sampler` `sampler`.
    The new solutions are added to the population.

    To improve performance it is advised to restrict the in general global
    search to a local one, e.g. by using Reverse Annealing.

    If `sampler` needs an embedding, `FixedEmbeddingComposite` or
    `LazyFixedEmbeddingComposite` are strongly advised to avoid recalculation in
    each generation.

    """
    def __init__(self, sampler: Sampler, sampler_config: Dict[str, Any] = {}):
        mutation = _mutation.ReverseAnnealingMutation(sampler, sampler_config)
        self._internal = _mutation.AdditionMutation(mutation)

    def initialise(self, problem: BinaryQuadraticModel) -> _mutation.Mutator:
        return self._internal.initialise(problem)


class RandomBitsRecombination(_recombination.RandomBitsRecombination):
    """
    Procreates new solutions by combining pairs of differing individuals. Each
    element of the old population is paired about `recombination_rate` times,
    eliminating duplications.

    For a single pair `num_bits` differing bits are randomly selected and
    flipped creating two offsprings that are added to the population.

    :ivar rng: A random number generator, that might be used for reproducibility.
    """
    def __init__(self, num_bits: int, recombination_rate: int, rng: Callable[[], Generator]):
        recombination = _recombination.RandomBitsRecombination(num_bits, recombination_rate, rng)
        self._internal = _recombination.AdditionRecombination(recombination)

    def initialise(self, problem: BinaryQuadraticModel) -> _recombination.Recombinator:
        return self._internal.initialise(problem)


class RandomClusterRecombination(_recombination.RandomClusterRecombination):
    """
    Procreates new solutions by combining pairs of differing individuals. Each
    element of the old population is paired about `recombination_rate` times,
    eliminating duplications.

    For a single pair a coherent region of differing bits is randomly selected and
    flipped creating two offsprings that are added to the population.

    :ivar rng: A random number generator, that might be used for reproducibility.
    """
    def __init__(self, recombination_rate: int, rng: Callable[[], Generator]):
        recombination = _recombination.RandomClusterRecombination(recombination_rate, rng)
        self._internal = _recombination.AdditionRecombination(recombination)

    def initialise(self, problem: BinaryQuadraticModel) -> _recombination.Recombinator:
        return self._internal.initialise(problem)


class TruncationSelection(_selection.TruncationSelection):
    """
    Selects the best `num_preserved` solutions, i.e. those with the lowest
    energies, and adds `num_new_states` random individuals. If the old
    population's size is less `num_preserved`, the deficit is compensated with
    random solutions, too.

    This helps to prevent so called elitism, i.e. the degeneration of the
    population of later generations because all individuals have been mutated to
    the same, possibly local, minima. New, unrelated solution widen the search
    space and increase the chance of finding global minima.

    :ivar rng: A random number generator, that might be used for reproducibility.
    """
    def __init__(self, num_preserved: int, num_new_solutions: int, rng: Callable[[], Generator]):
        assert num_preserved > 1
        assert num_new_solutions > 1
        recombination = _selection.TruncationSelection(num_preserved)
        self._internal = _selection.RefillingSelection(num_preserved + num_new_solutions, recombination, rng)

    def initialise(self, problem: BinaryQuadraticModel) -> _selection.Selector:
        return self._internal.initialise(problem)


class GenerationTermination(_termination.GenerationTermination):
    """
    Requests termination, after it has been called a `num_generations` times.

    It is used to restrict the number of generations, i.e. iterations, and force terminate.
    """


class LowestEnergyTermination(_termination.LowestEnergyTermination):
    """
    Requests termination, if a `predicate` is satisfied by the current
    populations's lowest energy.
    """


class AndTermination(_termination.AndTermination):
    """
    Combines multiple instances of `Terminator` and requests termination
    if all children do.
    """

class OrTermination(_termination.OrTermination):
    """
    Combines multiple instances of `Terminator` and requests termination
    if at least one of the children do.
    """



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dimod import SampleSet, BinaryQuadraticModel

from .startup import Startup

class WarmStartup(Startup):
    def __init__(self, population: SampleSet):
        self._population = population

    def startup(self, problem: BinaryQuadraticModel) -> SampleSet:
        assert self._population.variables == problem.variables
        assert self._population.vartype == problem.vartype

        return self._population



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dimod import SampleSet, BinaryQuadraticModel

from .recombination import Recombinator, Recombination


class IdentityRecombinator(Recombinator):
    """
    Passes the input population without any changes.
    """
    def recombine(self, population: SampleSet) -> SampleSet:
        return population


class IdentityRecombination(Recombination):
    def initialise(self, problem: BinaryQuadraticModel) -> Recombinator:
        return IdentityRecombinator()



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dimod import Sampler, SampleSet, BinaryQuadraticModel
from typing import Dict, Any

from .mutation import Mutator, Mutation


class ReverseAnnealingMutator(Mutator):
    """
    Improves the population by starting an annealing run at each individuals
    """
    def __init__(self, sampler: Sampler, bqm: BinaryQuadraticModel, sampler_config: Dict[str, Any] = {}):
        self._sampler = sampler
        self._sampler_config = sampler_config
        self._bqm = bqm

    def mutate(self, population: SampleSet) -> SampleSet:
        return self._sampler.sample(self._bqm, initial_states=population, **self._sampler_config)


class ReverseAnnealingMutation(Mutation):
    """
    A factory that produces instances `ReverseAnnealingMutation`.
    """
    def __init__(self, sampler: Sampler, sampler_config: Dict[str, Any] = {}):
        self._sampler = sampler
        self._sampler_config = sampler_config

    def initialise(self, problem: BinaryQuadraticModel) -> Mutator:
        return ReverseAnnealingMutator(self._sampler, problem, self._sampler_config)



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Callable
from dimod import BinaryQuadraticModel
from dimod.sampleset import SampleSet, concatenate
from numpy.random import Generator

from .selection import Selector, Selection
from ..startup.random_startup import create_random


MAX_FILLING_ITERATION = 1_000

class RefillingSelector(Selector):
    """
    Wraps an `internal` `Selector` and ensures that the next generation's
    population contains at least `num_preserved` individuals by adding random
    ones.

    This helps to prevent so called elitism, i.e. the degeneration of the
    population of later generations because all individuals have been mutated to
    the same, possibly local, minima. New, unrelated solution widen the search
    space and increase the chance of finding global minima.
    """
    def __init__(self, num_preserved: int, problem: BinaryQuadraticModel,
                 internal: Selector, rng: Generator):
        assert num_preserved > 1
        self._num_preserved = num_preserved
        self._bqm = problem
        self._internal = internal
        self._rng = rng

    def select(self, population: SampleSet) -> SampleSet:
        # Do not try to pick more states that available.
        new_population = self._internal.select(population)

        for _ in range(MAX_FILLING_ITERATION):
            if len(new_population) == self._num_preserved:
                return new_population

            # Compute how many additional elements must be created including the replanishing ones.
            num_missing = self._num_preserved - len(new_population)

            additional_individuals = create_random(len(self._bqm), num_missing,
                                                   self._bqm.vartype, MAX_FILLING_ITERATION, self._rng)
            sampleset = SampleSet.from_samples_bqm((additional_individuals, self._bqm.variables),
                                                   self._bqm)
            new_population = concatenate([new_population, sampleset]).aggregate()

        raise Exception("Tried to create new individuals but failed {MAX_FILLING_ITERATION}-times!")


class RefillingSelection(Selection):
    """
    A factory that produces instances `RefillingSelection`.
    """
    def __init__(self, num_preserved: int, internal: Selection, rng: Callable[[], Generator]):
        assert num_preserved > 1
        self._num_preserved = num_preserved
        self._internal = internal
        self._rng = rng

    def initialise(self, problem: BinaryQuadraticModel) -> Selector:
        return RefillingSelector(
            self._num_preserved,
            problem,
            self._internal.initialise(problem),
            self._rng()
            )



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dimod import SampleSet, BinaryQuadraticModel

from .termination import Terminator, Termination


class GenerationTerminator(Terminator):
    """
    Requests termination, after it has been called a `num_generations` times.

    It is used to restrict the number of generations, i.e. iterations, and force terminate.

    :ivar num_generations: Number of iterations after that the algorithm should terminate.
    """
    def __init__(self, num_generations: int):
        assert num_generations > 0
        self._num_generations = num_generations
        self._current_generation = 0

    def should_terminate(self, population: SampleSet) -> bool:
        self._current_generation += 1
        return self._current_generation > self._num_generations


class GenerationTermination(Termination):
    """
    A factory that produces instances `GenerationTerminator`.

    :ivar num_generations: Number of iterations after that an algorithm should terminate.
    """
    def __init__(self, num_generations: int):
        assert num_generations > 0
        self.num_generations = num_generations

    def initialise(self, problem: BinaryQuadraticModel) -> Terminator:
        """
        Instantiates a `GenerationTerminator` that stops after `self.num_generations` generations.

        :param problem: The problem that should be optimised in the current run. However, it is ignored here.
        :return: A a new `GenerationTerminator`.
        """
        return GenerationTerminator(self.num_generations)



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod

from dimod import SampleSet, BinaryQuadraticModel


class Terminator(ABC):
    """
    An interface for terminating an algorithm.

    Instances of subclasses are usually created once by a suitable :class:`Termination` factory at the start of an
    optimisation run. After each iteration step :meth:`should_terminate` is called
    to decide with respect to the current population and possibly a custom, encapsulated state if an algorithm should
    terminate. If so, iteration stops and the :class:`Terminator` is destroyed. Thus,
    each run should have its own :class:`Terminator` that is independent from any other.
    """
    @abstractmethod
    def should_terminate(self, population: SampleSet) -> bool:
        """
        Decides if the calling algorithm should stop iteration.

        :param population: The calling algorithm's current population.
        :return: :code:`True` to terminate, :code:`False` to continue.
        """
        pass


class Termination(ABC):
    """
    An interface that is used to configure the termination condition of a genetic algorithm.

    In fact, it is a factory that produces instances of the interface :class:`Terminator`.
    """
    @abstractmethod
    def initialise(self, problem: BinaryQuadraticModel) -> Terminator:
        """
        Instantiates and configures a implementors of the interface `Terminator`.

        :param problem: The problem that should be optimised in the current run.
        """
        pass



# Copyright (C) 2022 PlanQK
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass

from dimod import SampleSet, BinaryQuadraticModel
from toolz.curried import pipe

from ..mutation import Mutation
from ..recombination import Recombination
from ..selection import Selection
from ..startup import Startup
from ..termination import Termination


@dataclass
class GeneticAlgorithm:
    """
    A configurable implementation of a genetic algorithm, that solves a minimisation problem given as a
    :class:`QuadraticModel` and returns a solution as :class:`SampleSet`. It is thus fully compatible with the
    `dimod`-API.

    This class itself is just a template that must be filled with the following components to work:

    :param Startup startup: Determines the initial population. For instance, it can be random or used as a warm start.

    :param Mutation mutation: Improves quality of the current population by starting a local search from its individuals.

    :param Recombination recombination: Procreates new individuals by combining existing ones.

    :param Selection selection: Truncates low quality solutions to keep the population small.

    :param Termination termination: Allows to stop the algorithm if the population has achieved a certain quality
    level or takes too long.

    Note that this class is stateless, i.e. calling `optimise` multiple times parallelly works without any problems.
    """
    startup: Startup
    mutation: Mutation
    recombination: Recombination
    selection: Selection
    termination: Termination

    def optimise(self, problem: BinaryQuadraticModel) -> SampleSet:
        # Instantiate each component for a each optimisation run. Therefore, they have their own internal state that
        # should not interfere with the outside environment.
        mutator = self.mutation.initialise(problem)
        recombinator = self.recombination.initialise(problem)
        selector = self.selection.initialise(problem)
        terminator = self.termination.initialise(problem)

        # Produce a first population
        population = self.startup.startup(problem)
        while True:
            # Add new individuals by mutation and recombination.
            population = pipe(population,
                              mutator.mutate,
                              recombinator.recombine)
            # Check termination condition, e.g. existence of a good solution.
            if terminator.should_terminate(population):
                break

            # Truncate low quality individuals.
            population = selector.select(population)
        return population





from abc import ABC, abstractmethod

from dimod import SampleSet, BinaryQuadraticModel


class Selector(ABC):
    """
    An interface for collocating the population of the succeeding iteration.

    Instances of subclasses are usually created once by a suitable :class:`Selector` factory at the start of an
    optimisation run. In each iteration step :meth:`select` is called to assess which individuals of the current
    population are eligible to be preserved for the next iteration step and, potentially, to add completely new,
    e.g. random, ones.

    When iteration terminates, the :class:`Selector` is destroyed. Thus, each run should have its own
    :class:`Selector` that is independent from any other.
    """
    @abstractmethod
    def select(self, population: SampleSet) -> SampleSet:
        """
        Composes the population of the succeeding iteration step based on the current population.

        Implementations of this interface must ensure that in the result population no solution appears twice.
        Moreover, the best individual of the new population must be at least as good as the best one of the old population.

        :param population: The calling algorithm's current population.
        """
        pass


class Selection(ABC):
    """
    An interface that is used to configure, how to select the individuals for the next generation.

    In fact, it is a factory that produces instances of the interface :class:`Selector`.
    """
    @abstractmethod
    def initialise(self, problem: BinaryQuadraticModel) -> Selector:
        """
        Instantiates and configures a implementors of the interface `Selector`.

        :param problem: The problem that should be optimised in the current run.
        """
        pass
